/* Xiao -- An outliner library for GTK
 * Copyright (C) 2002 Gary Benson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "xiaohandle.h"
#include <gtk/gtksignal.h>

enum {
    FOLD_EVENT,
    SELECT_EVENT,
    LAST_SIGNAL
};

static void	    xiao_handle_class_init	(XiaoHandleClass *class);
static void	    xiao_handle_init		(XiaoHandle *handle);
static void	    xiao_handle_size_request	(GtkWidget      *widget,
						 GtkRequisition *requisition);
static void	    xiao_handle_size_allocate	(GtkWidget     *widget,
						 GtkAllocation *allocation);
static void	    xiao_handle_realize		(GtkWidget *widget);
static gboolean	    xiao_handle_expose		(GtkWidget      *widget,
						 GdkEventExpose *event);
static gboolean	    xiao_handle_button_press	(GtkWidget      *widget,
						 GdkEventButton *event);
static void	    xiao_handle_force_redraw	(XiaoHandle *handle);


static gint handle_signals[LAST_SIGNAL] = { 0 };


GType
xiao_handle_get_type(void)
{
    static GType handle_type = 0;

    if(!handle_type) {
	static const GTypeInfo handle_info =
	{
	    sizeof(XiaoHandleClass),
	    NULL,		/* base_init */
	    NULL,		/* base_finalize */
	    (GClassInitFunc)xiao_handle_class_init,
	    NULL,		/* class_finalize */
	    NULL,		/* class_data */
	    sizeof(XiaoHandle),
	    0,			/* n_preallocs */
	    (GInstanceInitFunc)xiao_handle_init,
	};

      handle_type = g_type_register_static(
	  GTK_TYPE_WIDGET, "XiaoHandle", &handle_info, 0);
    }

    return handle_type;
}

static void
xiao_handle_class_init(XiaoHandleClass *class)
{
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;

    object_class = (GtkObjectClass *)class;
    widget_class = (GtkWidgetClass *)class;

    widget_class->size_request         = xiao_handle_size_request;
    widget_class->size_allocate        = xiao_handle_size_allocate;
    widget_class->realize              = xiao_handle_realize;
    widget_class->expose_event         = xiao_handle_expose;
    widget_class->button_press_event   = xiao_handle_button_press;
    
    handle_signals[FOLD_EVENT] =
	gtk_signal_new ("fold_event",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET(XiaoHandleClass, fold_event),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);    

    handle_signals[SELECT_EVENT] =
	gtk_signal_new ("select_event",
			GTK_RUN_FIRST,
			GTK_CLASS_TYPE(object_class),
			GTK_SIGNAL_OFFSET(XiaoHandleClass, select_event),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);    
}

static void
xiao_handle_init(XiaoHandle *handle)
{
    handle->folded = FALSE;
    handle->selected = FALSE;
}

GtkWidget*
xiao_handle_new(void)
{
    return GTK_WIDGET(g_object_new(xiao_handle_get_type(), NULL));
}

static void 
xiao_handle_size_request(GtkWidget      *widget,
			 GtkRequisition *requisition)
{
    PangoContext *context;
    PangoFontMetrics *metrics;
    gint ascent, descent;

    /* FIXME: you should probably call x_h_set_size() on creation,
     * which would set the size to something sensible based on the
     * parameters of the GtkTextView in use on the other side of the
     * XiaoItem (ergo all of this Pango crap can go).
     */
    
    g_return_if_fail(widget != NULL);
    g_return_if_fail(XIAO_IS_HANDLE(widget));
    g_return_if_fail(requisition != NULL);

    context = gtk_widget_get_pango_context(widget);
    metrics = pango_context_get_metrics(context,
					widget->style->font_desc,
					pango_context_get_language(context));

    ascent  = pango_font_metrics_get_ascent(metrics);
    descent = pango_font_metrics_get_descent(metrics);
    
    requisition->height = PANGO_PIXELS(ascent + descent);
    requisition->width  = requisition->height;

    pango_font_metrics_unref (metrics);
}

static void
xiao_handle_size_allocate(GtkWidget     *widget,
			  GtkAllocation *allocation)
{
    XiaoHandle *handle;

    g_return_if_fail(widget != NULL);
    g_return_if_fail(XIAO_IS_HANDLE(widget));
    g_return_if_fail(allocation != NULL);

    widget->allocation = *allocation;
    handle = XIAO_HANDLE (widget);

    if(GTK_WIDGET_REALIZED(widget)) {
	gdk_window_move_resize(widget->window,
			       allocation->x, allocation->y,
			       allocation->width, allocation->height);

    }

    handle->width  = allocation->width;
    handle->height = allocation->height;
}

static void
xiao_handle_realize(GtkWidget *widget)
{
    XiaoHandle *handle;
    GdkWindowAttr attributes;
    gint attributes_mask;

    g_return_if_fail(widget != NULL);
    g_return_if_fail(XIAO_IS_HANDLE(widget));

    GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);
    handle = XIAO_HANDLE(widget);

    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.event_mask = gtk_widget_get_events(widget) | 
	GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK;
    attributes.visual = gtk_widget_get_visual(widget);
    attributes.colormap = gtk_widget_get_colormap(widget);

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    widget->window = gdk_window_new(
	widget->parent->window, &attributes, attributes_mask);

    widget->style = gtk_style_attach(widget->style, widget->window);

    gdk_window_set_user_data(widget->window, widget);

    gdk_window_set_background(
	widget->window, &widget->style->base[GTK_STATE_NORMAL]);
}

static gboolean
xiao_handle_expose(GtkWidget      *widget,
		   GdkEventExpose *event)
{
    XiaoHandle *handle;
    gint x, y, w, h, hw, hh, i, j;
    GtkStateType fg_state, bg_state;
    GdkGC *fg, *bg;
    
    g_return_val_if_fail(widget != NULL, FALSE);
    g_return_val_if_fail(XIAO_IS_HANDLE(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    if (event->count > 0)
	return FALSE;
  
    handle = XIAO_HANDLE(widget);

    /* Calculate the dimensions of the arrow.
     * Based on gtkstyle.c (calculate_arrow_geometry), but modified
     * since we want to shrink the arrow so it doesn't dwarf the text
     * next to it and yet ensure it looks good and remains big enough
     * to click when small fonts are used.
     */
    x = y = 0;
    hw = handle->width;
    hh = handle->height;
	
    h = hh / 2;         /* ours is only half the height of the box... */
    h += (h % 2) + 1;   /* ...but we grow to fit instead of shrinking */
    w = (h / 2 + 1);

    if (w > hw) {
	w = hw;
	h = 2 * w - 1;
    }
    if (hw % 2 == 0 || w % 2 == 0)
	hw -= 1;
    
    x += (hw - w) / 2;
    y += (hh - h) / 2;
    y += (hh - h) % 2;  /* float down rather than up when not symmetrical
			   (works better with small fonts) */

    /* Draw it */
    bg_state = handle->selected ? GTK_STATE_SELECTED : GTK_STATE_NORMAL;
    fg_state = handle->folded ? bg_state : GTK_STATE_INSENSITIVE;
    bg = widget->style->base_gc[bg_state];
    fg = widget->style->text_gc[fg_state];
    
    gdk_draw_rectangle(
	widget->window, bg, TRUE, 0, 0, handle->width, handle->height);
    
    /* nicked from gtkstyle.c (draw_arrow) */
    for (i = 0, j = 0; i < w; i++, j++)
	gdk_draw_line(widget->window, fg, x + i, y + j, x + i, y + h - j - 1);

    return FALSE;
}

static gboolean
xiao_handle_button_press(GtkWidget      *widget,
			 GdkEventButton *event)
{
    XiaoHandle *handle;

    g_return_val_if_fail(widget != NULL, FALSE);
    g_return_val_if_fail(XIAO_IS_HANDLE(widget), FALSE);
    g_return_val_if_fail(event != NULL, FALSE);

    handle = XIAO_HANDLE(widget);

    if (event->type == GDK_BUTTON_PRESS && event->button == 1) {
	gtk_signal_emit(GTK_OBJECT(handle), handle_signals[SELECT_EVENT]);
	return TRUE;
    }
    if (event->type == GDK_2BUTTON_PRESS && event->button == 1) {
	/* NB: a 2BUTTON_PRESS is always preceded by two BUTTON_PRESSes */
	gtk_signal_emit(GTK_OBJECT(handle), handle_signals[FOLD_EVENT]);
	return TRUE;
    }
    if (event->type == GDK_BUTTON_PRESS && event->button == 2) {
	gtk_signal_emit(GTK_OBJECT(handle), handle_signals[FOLD_EVENT]);
	return TRUE;
    }
    if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
	/* FIXME: insert popup menu here */
    }
    
    return FALSE;
}

void
xiao_handle_set_folded(XiaoHandle *handle, gboolean folded)
{
    g_return_if_fail(handle != NULL);
    g_return_if_fail(XIAO_IS_HANDLE(handle));

    if (handle->folded == folded)
	return;

    handle->folded = folded;
    xiao_handle_force_redraw(handle);
}

gboolean
xiao_handle_get_folded(XiaoHandle *handle)
{
    g_return_val_if_fail(handle != NULL, FALSE);
    g_return_val_if_fail(XIAO_IS_HANDLE(handle), FALSE);

    return handle->folded;
}

void
xiao_handle_set_selected(XiaoHandle *handle, gboolean selected)
{
    g_return_if_fail(handle != NULL);
    g_return_if_fail(XIAO_IS_HANDLE(handle));

    if (handle->selected == selected)
	return;

    handle->selected = selected;
    xiao_handle_force_redraw(handle);
}

gboolean
xiao_handle_get_selected(XiaoHandle *handle)
{
    g_return_val_if_fail(handle != NULL, FALSE);
    g_return_val_if_fail(XIAO_IS_HANDLE(handle), FALSE);

    return handle->selected;
}

static void
xiao_handle_force_redraw(XiaoHandle *handle)
{
    GdkRectangle rect;
    
    g_return_if_fail(handle != NULL);
    g_return_if_fail(XIAO_IS_HANDLE(handle));

    rect.x = 0;
    rect.y = 0;
    rect.width  = handle->width;
    rect.height = handle->height;
    
    gdk_window_invalidate_rect(GTK_WIDGET(handle)->window, &rect, FALSE);
}
